/*====================================================================
| Version: November 11, 2010
\===================================================================*/

/*====================================================================
| Ricard Delgado-Gonzalo
| EPFL/STI/IOA/LIB/BM.4.141
| e-mail: ricard.delgado@gmail.com
| URL: http://www.ricarddelgado.com/
\===================================================================*/
import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.Macro;
import ij.gui.GenericDialog;
import ij.gui.Roi;
import ij.plugin.filter.ExtendedPlugInFilter;
import ij.plugin.filter.PlugInFilterRunner;
import ij.plugin.frame.Recorder;
import ij.process.FloatProcessor;
import ij.process.ImageProcessor;
import io.CTInput;
import io.CTOutput;

import java.awt.Checkbox;
import java.awt.TextField;
import java.util.Vector;

import display.CTLUT;
import display.CTOverlay;
import display.CTOverlay.CTCanvas;

import segmentation.CTSegmentation;
import structures.CTConnectivity;
import structures.CTExtractedData;
import structures.CTRunParameters;
import tracking.CTTracking;


public class CellTracker_Cluster implements ExtendedPlugInFilter{
	//interface
	private final GenericDialog dialog = new GenericDialog("Cell Tracker - Cluster Version");
	
	private static final String XCORNERBOUNDINGBOXLABEL 	= "ROI_X_corner";
	private static final String YCORNERBOUNDINGBOXLABEL 	= "ROI_Y_corner";
	private static final String WIDTHBOUNDINGBOXLABEL 		= "ROI_width";
	private static final String HEIGHTBOUNDINGBOXLABEL 		= "ROI_height";
	private static final String REGIONOFFSETLABEL 			= "Region_offset";
	private static final String SIGMANONMOVINGREGIONLABEL	= "Threshold_NMR";
	private static final String SIGMALABEL					= "Smoothing_std";
	private static final String LAMBDALABEL					= "N-regularization";
	private static final String UPDISTANCELABEL				= "Max_up_displacement";
	private static final String DOWNDISTANCELABEL			= "Max_down_displacement";
	private static final String LATERALDISTANCELABEL		= "Max_lateral displacement";
	private static final String REGULARIZATIONITERLABEL		= "N-regularization_iter";
	private static final String INPUTFOLDERCH1LABEL			= "Input_folder_1st_channel";
	private static final String INPUTFOLDERCH2LABEL			= "Input_folder_2nd_channel";
	private static final String OUTPUTFILELABEL 			= "Output_file";
	private static final String DISABLETRACKINGLABEL		= "Disable_tracking";
	private static final String VERBOSELABEL				= "Verbose";
	
	//default values
	private static final int XCORNERBOUNDINGBOXDEFAULT 		= 0;
	private static final int YCORNERBOUNDINGBOXDEFAULT 		= 0;
	private static final int WIDTHBOUNDINGBOXDEFAULT 		= 512;
	private static final int HEIGHTBOUNDINGBOXDEFAULT 		= 512;
	private static final int REGIONOFFSETDEFAULT			= 3;
	private static final double SIGMANONMOVINGREGIONDEFAULT = 1;
	private static final double SIGMADEFAULT				= 4.0;
	private static final double LAMBDADEFAULT				= 0;
	private static final double UPDISTANCEDEFAULT			= 30;
	private static final double DOWNDISTANCEDEFAULT			= 10;
	private static final double LATERALDISTANCEDEFAULT		= 10;
	private static final int REGULARIZATIONITERDEFAULT		= 0;
	private static final String	INPUTFOLDERCH1DEFAULT		= "";
	private static final String	INPUTFOLDERCH2DEFAULT		= "";
	private static final String	OUTPUTFILEDEFAULT 			= "";
	private static final boolean DISABLETRACKINGDEFAULT		= false;
	private static final boolean VERBOSEDEFAULT				= true;
	
	private String inputFolderCh1 = null;
	private String inputFolderCh2 = null;
	private String[] fileListCh1 = null;
	private String[] fileListCh2 = null;
	private String outputFile;
	
	//execution parameters
	private static CTRunParameters runParameters = null;
	private int nFiles;
	private final static int CAPABILITIES = NO_IMAGE_REQUIRED;

	private static boolean disableTracking					= DISABLETRACKINGDEFAULT;
	private static boolean verbose							= VERBOSEDEFAULT;
	
	//extracted data
	private CTExtractedData[] extractedData = null;
	private CTConnectivity[] linkedData = null;
	
	//auxiliar classes
	CTOverlay overlayCh1;
	CTOverlay overlayCh2;
	private ImagePlus imCh1 = null;
	private ImagePlus imCh2 = null;
	
	@SuppressWarnings("rawtypes")
	public void run (ImageProcessor ip) {
		//Loading interface parameters
		final Vector numbers 	= dialog.getNumericFields();
		final Vector strings 	= dialog.getStringFields();
		final Vector checkboxes = dialog.getCheckboxes();
		runParameters = new CTRunParameters();
		final int xCornerBoundingBox		= (new Integer(((TextField)numbers.elementAt(0)).getText())).intValue();
		final int yCornerBoundingBox		= (new Integer(((TextField)numbers.elementAt(1)).getText())).intValue();
		final int widthBoundingBox 			= (new Integer(((TextField)numbers.elementAt(2)).getText())).intValue();
		final int heightBoundingBox			= (new Integer(((TextField)numbers.elementAt(3)).getText())).intValue();
		runParameters.regionOffset			= (new Integer(((TextField)numbers.elementAt(4)).getText())).intValue();
		runParameters.sigmaNotMovingRegion	= (new Double(((TextField)numbers.elementAt(5)).getText())).doubleValue();
		runParameters.sigma					= (new Double(((TextField)numbers.elementAt(6)).getText())).doubleValue();
		runParameters.lambda				= (new Double(((TextField)numbers.elementAt(7)).getText())).doubleValue();
		runParameters.upDisplacement		= (new Double(((TextField)numbers.elementAt(8)).getText())).doubleValue();
		runParameters.downDisplacement		= (new Double(((TextField)numbers.elementAt(9)).getText())).doubleValue();
		runParameters.lateralDisplacement	= (new Double(((TextField)numbers.elementAt(10)).getText())).doubleValue();
		runParameters.regularizationIter	= (new Integer(((TextField)numbers.elementAt(11)).getText())).intValue();
		inputFolderCh1 						= ((TextField)strings.elementAt(0)).getText();
		inputFolderCh2 						= ((TextField)strings.elementAt(1)).getText();
		outputFile 							= ((TextField)strings.elementAt(2)).getText();
		disableTracking 					= ((Checkbox)checkboxes.elementAt(0)).getState();
		verbose 							= ((Checkbox)checkboxes.elementAt(1)).getState();
		
		Roi roi = new Roi(xCornerBoundingBox, yCornerBoundingBox, widthBoundingBox, heightBoundingBox);
		runParameters.roiBox = roi.getBounds();

		if(inputFolderCh1.length()==0 || inputFolderCh2.length()==0) {
			IJ.error("No destination file specified");
			return;
		}

		if(outputFile.length()==0) {
			IJ.error("No destination file specified");
			return;
		}
	
		Recorder.setCommand("CellTracker Cluster");
		Recorder.recordOption(XCORNERBOUNDINGBOXLABEL,""+xCornerBoundingBox);
		Recorder.recordOption(YCORNERBOUNDINGBOXLABEL,""+yCornerBoundingBox);
		Recorder.recordOption(WIDTHBOUNDINGBOXLABEL,""+widthBoundingBox);
		Recorder.recordOption(HEIGHTBOUNDINGBOXLABEL,""+heightBoundingBox);
		Recorder.recordOption(REGIONOFFSETLABEL,""+runParameters.regionOffset);
		Recorder.recordOption(SIGMANONMOVINGREGIONLABEL,""+runParameters.sigmaNotMovingRegion);
		Recorder.recordOption(SIGMALABEL,""+runParameters.sigma);
		Recorder.recordOption(LAMBDALABEL,""+runParameters.lambda);
		Recorder.recordOption(UPDISTANCELABEL,""+runParameters.upDisplacement);
		Recorder.recordOption(DOWNDISTANCELABEL,""+runParameters.downDisplacement);
		Recorder.recordOption(LATERALDISTANCELABEL,""+runParameters.lateralDisplacement);
		Recorder.recordOption(REGULARIZATIONITERLABEL,""+runParameters.regularizationIter);
		Recorder.recordOption(INPUTFOLDERCH1LABEL,inputFolderCh1);
		Recorder.recordOption(INPUTFOLDERCH2LABEL,inputFolderCh2);
		Recorder.recordOption(OUTPUTFILELABEL,outputFile);
		Recorder.recordOption(DISABLETRACKINGLABEL,""+disableTracking);
		Recorder.recordOption(VERBOSELABEL,""+verbose);
		Recorder.saveCommand();
				
		fileListCh1 = CTInput.getListDir(inputFolderCh1);
		if(fileListCh1==null)	return;

		fileListCh2 = CTInput.getListDir(inputFolderCh2);
		if(fileListCh2==null)	return;

		if(fileListCh1.length!=fileListCh2.length){
			IJ.error("Stack sizes do not match");
			return;
		}
		
		nFiles = fileListCh1.length;
		
		//checking ROI
		FloatProcessor ipr = CTInput.readProcessor(inputFolderCh1,fileListCh1[0], null);
		int imWidth = ipr.getWidth();
		int imHeight = ipr.getHeight();
		
		if(xCornerBoundingBox>=imWidth || yCornerBoundingBox>=imHeight){
			IJ.error("Selected ROI is outside the imaging area");
			return;
		}
		
		if(xCornerBoundingBox+widthBoundingBox>=imWidth){
			writeLog("Selected ROI is partially outside the imaging area...");
			writeLog("Cropping ROI");
			runParameters.roiBox.width = imWidth-xCornerBoundingBox;
		}
		
		if(yCornerBoundingBox+heightBoundingBox>=imHeight){
			writeLog("Selected ROI is partially outside the imaging area...");
			writeLog("Cropping ROI");
			runParameters.roiBox.height = imHeight-yCornerBoundingBox;
		}
		
		boolean success = runSegmentation();
		if(success){
			runTracking();
			CTOutput.saveXML(outputFile, runParameters, extractedData, linkedData);
			CTOutput.saveSingleCell(outputFile, extractedData);
			CTOutput.saveSingleCellOutline(outputFile, extractedData);
			if(verbose) displayResults(roi = new Roi(xCornerBoundingBox, yCornerBoundingBox, runParameters.roiBox.width, runParameters.roiBox.height));
		}

	} 
	
	private void runTracking(){
		writeLog("Starting linking stage...");
		linkedData = new CTConnectivity[nFiles];
		for(int frame=0; frame<nFiles; frame++){
			linkedData[frame] = new CTConnectivity(extractedData[frame].getNregions());
			linkedData[frame].frame = frame;
		}
		if(!disableTracking){
			CTTracking tracking = new CTTracking(runParameters, extractedData, linkedData);
			tracking.doTracking();
		}
		writeLog("Linking stage finished");
	}
	
	private boolean runSegmentation(){
		writeLog("Starting detection and segmentation stage...");
		extractedData = new CTExtractedData[nFiles];
		CTSegmentation segmentation = new CTSegmentation(inputFolderCh1, fileListCh1, inputFolderCh2, fileListCh2, runParameters, extractedData);
		segmentation.showWatershedRegions(verbose);
		segmentation.showTempStd(verbose);
		segmentation.showNMRMap(verbose);
		segmentation.doSegmentation();
		writeLog("Detection and segmentation stage finished");
		return segmentation.isSuccessful();
	}
	
	
	public void setNPasses (int nPasses) {
	}
	
	public int setup (final String arg, final ImagePlus imp) {
		return(CAPABILITIES);
	} 
	
	public int showDialog (final ImagePlus imp, final String command, final PlugInFilterRunner pfr) {
		dialog.addNumericField(XCORNERBOUNDINGBOXLABEL, XCORNERBOUNDINGBOXDEFAULT, 0);
		dialog.addNumericField(YCORNERBOUNDINGBOXLABEL, YCORNERBOUNDINGBOXDEFAULT, 0);
		dialog.addNumericField(WIDTHBOUNDINGBOXLABEL, WIDTHBOUNDINGBOXDEFAULT, 0);
		dialog.addNumericField(HEIGHTBOUNDINGBOXLABEL, HEIGHTBOUNDINGBOXDEFAULT, 0);
		dialog.addNumericField(REGIONOFFSETLABEL, REGIONOFFSETDEFAULT, 0);
		dialog.addNumericField(SIGMANONMOVINGREGIONLABEL, SIGMANONMOVINGREGIONDEFAULT, 3);
		dialog.addNumericField(SIGMALABEL, SIGMADEFAULT, 1);
		dialog.addNumericField(LAMBDALABEL, LAMBDADEFAULT, 0);
		dialog.addNumericField(UPDISTANCELABEL, UPDISTANCEDEFAULT, 1);
		dialog.addNumericField(DOWNDISTANCELABEL, DOWNDISTANCEDEFAULT, 1);
		dialog.addNumericField(LATERALDISTANCELABEL, LATERALDISTANCEDEFAULT, 1);
		dialog.addNumericField(REGULARIZATIONITERLABEL, REGULARIZATIONITERDEFAULT, 0);
		dialog.addStringField(INPUTFOLDERCH1LABEL, INPUTFOLDERCH1DEFAULT);
		dialog.addStringField(INPUTFOLDERCH2LABEL, INPUTFOLDERCH2DEFAULT);
		dialog.addStringField(OUTPUTFILELABEL, OUTPUTFILEDEFAULT);
		dialog.addCheckbox(DISABLETRACKINGLABEL, DISABLETRACKINGDEFAULT);
		dialog.addCheckbox(VERBOSELABEL, VERBOSEDEFAULT);
		dialog.addPanel(new CTCreditsButton());
		
		if (Macro.getOptions() != null) {
			activateMacro(imp);
			return(CAPABILITIES);
		}
		dialog.showDialog();
		if (dialog.wasCanceled()) {
			return(DONE);
		}
		if (dialog.wasOKed()) {
			return(CAPABILITIES);
		}else {
			return(DONE);
		}
	} 
	

	@SuppressWarnings("rawtypes")
	private void activateMacro (final ImagePlus imp) {
		final Vector numbers 		= dialog.getNumericFields();
		final Vector strings 		= dialog.getStringFields();
		final Vector checkboxes 	= dialog.getCheckboxes();
	
		final TextField xCornerBoundingBoxTextField 	= (TextField)numbers.elementAt(0);
		final TextField yCornerBoundingBoxTextField 	= (TextField)numbers.elementAt(1);
		final TextField widthBoundingBoxTextField 		= (TextField)numbers.elementAt(2);
		final TextField heightBoundingBoxTextField		= (TextField)numbers.elementAt(3);
		final TextField regionOffsetTextField			= (TextField)numbers.elementAt(4);
		final TextField sigmaNonMovingRegionTextField	= (TextField)numbers.elementAt(5);
		final TextField sigmaTextField					= (TextField)numbers.elementAt(6);
		final TextField lambdaTextField					= (TextField)numbers.elementAt(7);
		final TextField maxUpDistanceTextField			= (TextField)numbers.elementAt(8);
		final TextField maxDownDistanceTextField		= (TextField)numbers.elementAt(9);
		final TextField maxLateralDistanceTextField		= (TextField)numbers.elementAt(10);
		final TextField regularizationTextField			= (TextField)numbers.elementAt(11);
		final TextField inputFolderCh1TextField 		= (TextField)strings.elementAt(0);
		final TextField inputFolderCh2TextField 		= (TextField)strings.elementAt(1);
		final TextField outputFileTextField   			= (TextField)strings.elementAt(2);
		final Checkbox disableTrackingCheckBox			= (Checkbox)checkboxes.elementAt(0);
		final Checkbox verboseCheckBox					= (Checkbox)checkboxes.elementAt(1);
		
		final String options = Macro.getOptions();
		
		xCornerBoundingBoxTextField.setText(Macro.getValue(options, XCORNERBOUNDINGBOXLABEL, ""+XCORNERBOUNDINGBOXDEFAULT));
		yCornerBoundingBoxTextField.setText(Macro.getValue(options, YCORNERBOUNDINGBOXLABEL, ""+YCORNERBOUNDINGBOXDEFAULT));
		widthBoundingBoxTextField.setText(Macro.getValue(options, WIDTHBOUNDINGBOXLABEL, ""+WIDTHBOUNDINGBOXDEFAULT));
		heightBoundingBoxTextField.setText(Macro.getValue(options, HEIGHTBOUNDINGBOXLABEL, ""+HEIGHTBOUNDINGBOXDEFAULT));
		regionOffsetTextField.setText(Macro.getValue(options, REGIONOFFSETLABEL, ""+REGIONOFFSETDEFAULT));
		sigmaNonMovingRegionTextField.setText(Macro.getValue(options, SIGMANONMOVINGREGIONLABEL, ""+SIGMANONMOVINGREGIONDEFAULT));
		sigmaTextField.setText(Macro.getValue(options, SIGMALABEL, ""+SIGMADEFAULT));
		lambdaTextField.setText(Macro.getValue(options, LAMBDALABEL, ""+LAMBDADEFAULT));
		maxUpDistanceTextField.setText(Macro.getValue(options, UPDISTANCELABEL, ""+UPDISTANCEDEFAULT));
		maxDownDistanceTextField.setText(Macro.getValue(options, DOWNDISTANCELABEL, ""+DOWNDISTANCEDEFAULT));
		maxLateralDistanceTextField.setText(Macro.getValue(options, LATERALDISTANCELABEL, ""+LATERALDISTANCEDEFAULT));
		regularizationTextField.setText(Macro.getValue(options, REGULARIZATIONITERLABEL, ""+REGULARIZATIONITERDEFAULT));
		inputFolderCh1TextField.setText(Macro.getValue(options, INPUTFOLDERCH1LABEL, ""+INPUTFOLDERCH1DEFAULT));
		inputFolderCh2TextField.setText(Macro.getValue(options, INPUTFOLDERCH2LABEL, ""+INPUTFOLDERCH2DEFAULT));
		outputFileTextField.setText(Macro.getValue(options, OUTPUTFILELABEL, ""+OUTPUTFILEDEFAULT));
		
		String s1 = new String(Macro.getValue(options, DISABLETRACKINGLABEL, ""+disableTracking));
		if(s1.equals("true")){
			disableTrackingCheckBox.setState(true);
		}else{
			disableTrackingCheckBox.setState(false);
		}
		
		String s2 = new String(Macro.getValue(options, VERBOSELABEL, ""+verbose));
		if(s2.equals("true")){
			verboseCheckBox.setState(true);
		}else{
			verboseCheckBox.setState(false);
		}
	} 
	
	private void displayResults(Roi roi){
		loadStack(roi);
		overlayCh1 = new CTOverlay(imCh1, extractedData, linkedData);
		overlayCh2 = new CTOverlay(imCh2, extractedData, linkedData);
		overlayCh1.setOtherChannel(overlayCh2);
		overlayCh2.setOtherChannel(overlayCh1);
		overlayCh1.canvas.setEditMode(CTCanvas.IDLESTATE);
		overlayCh2.canvas.setEditMode(CTCanvas.IDLESTATE);
		overlayCh1.canvas.showCellOutline(true);
		overlayCh2.canvas.showCellOutline(true);
		overlayCh1.canvas.setDisplayMode(CTCanvas.TRACKMODE);
		overlayCh2.canvas.setDisplayMode(CTCanvas.TRACKMODE);
		
		
	}
	
	private void loadStack(Roi roi){
		FloatProcessor fpch1 = CTInput.readProcessor(inputFolderCh1, fileListCh1[0], roi);
		FloatProcessor fpch2 = CTInput.readProcessor(inputFolderCh2, fileListCh2[0], roi);

		CTLUT.setLUT(fpch2, CTLUT.GREEN);
		
		int nx, ny;
		if(roi==null){
			nx = fpch1.getWidth();
			ny = fpch1.getHeight();
		}else{
			nx = roi.getBounds().width;
			ny = roi.getBounds().height;
		}
		
		ImageStack originalStackCh1 = new ImageStack(nx, ny);
		ImageStack originalStackCh2 = new ImageStack(nx, ny);
		originalStackCh1.addSlice("", fpch1);
		originalStackCh2.addSlice("", fpch2);
	
		for(int frame=1; frame<nFiles; frame++){
			fpch1 = CTInput.readProcessor(inputFolderCh1, fileListCh1[frame], roi);
			fpch2 = CTInput.readProcessor(inputFolderCh2, fileListCh2[frame], roi);
			CTLUT.setLUT(fpch2, CTLUT.GREEN);
			originalStackCh1.addSlice("", fpch1);
			originalStackCh2.addSlice("", fpch2);
		}
		
		writeLog("Stack of "+ nFiles + " files loaded");
	
		imCh1 = new ImagePlus("Segmentation Window - Channel 1", originalStackCh1);
		imCh2 = new ImagePlus("Segmentation Window - Channel 2", originalStackCh2);
		imCh1.show();
		imCh2.show();
	}
	
	private void writeLog(String str){
		if(verbose) IJ.log(str);
	}
	
} 