/*====================================================================
| Version: December 8, 2010
\===================================================================*/

/*====================================================================
| Ricard Delgado-Gonzalo
| EPFL/STI/IOA/LIB/BM.4.141
| e-mail: ricard.delgado@gmail.com
| URL: http://www.ricarddelgado.com/
\===================================================================*/
package filters;
import ij.ImagePlus;
import ij.process.ColorProcessor;
import ijtools.Convolver2D;
import ijtools.IJmath;
import ijtools.IJtools;
import ijtools.Interpolator;

public class CTSteerableDetector{

	private double sigma;
	
	private double a20;
	private double a22;
	private double a40;
	private double a42;
	private double a44;
	
	private double a11;
	private double a31;
	private double a33;
	private double a51;
	private double a53;
	
	private double[] alpha;
	
	private double[] gx;
	private double[] gy;
	
	private double[] gxx;
	private double[] gxy;
	private double[] gyy;
	
	private double[] gxxx;
	private double[] gxxy;
	private double[] gxyy;
	private double[] gyyy;
	
	private double[] gxxxx;
	private double[] gxxxy;
	private double[] gxxyy;
	private double[] gxyyy;
	private double[] gyyyy;
	
	private double[] gxxxxx;
	private double[] gxxxxy;
	private double[] gxxxyy;
	private double[] gxxyyy;
	private double[] gxyyyy;
	private double[] gyyyyy;
	
	private double[] input, response, orientation;
	private int nx, ny, nz, size, nxy;
	private int M;
	
	private static final double PI = Math.PI;
	private static final double tolerance = 1e-13;
	
	private boolean stop = false;
	
		
	/**
	* Initializes a new feature detector.
	*
	* @param input The input image or stack. Its dimensions are passed as nx, ny, nz.
	* @param sigma The standard deviation of the Gaussian on which the steerable templates are based. This controls the feature width.
	* @param M The order of the feature template. Order up to 5 are supported in the current implementation.
	* @param alpha The array of weights that control the linear combination of Gaussian derivatives up to order M which constitute the feature template. For more information, see the documentation for the ParameterSet class. 
	*/
	public CTSteerableDetector(double[] input, int nx, int ny, int nz, double sigma, int M, double[] alpha) {
		this.input = input;
		this.M = M;
		this.sigma = sigma;
		this.alpha = alpha;
		this.nx = nx;
		this.ny = ny;
		this.nz = nz;
		
		nxy = nx*ny;
		size = nxy*nz;
		
		response = new double[size];
		orientation = new double[size];
	}
	
	public void stop() {
		stop = true;
	}
	
	public boolean getStop() {
		return stop;
	}
	
	
	/**
	* Computes the feature detection. The results are stored in "response" and "orientation".
	*/
	public void filter() {
		switch(M) {
			case 1:
				a11 = alpha[0];
				filterM1();
				break;
			case 2:
				a20 = alpha[0];
				a22 = alpha[1];
				filterM2xx();
				break;
			case 3:
				a11 = alpha[0];
				a31 = alpha[1];
				a33 = alpha[2];
				filterM3();
				break;
			case 4:
				a20 = alpha[0];
				a22 = alpha[1];
				a40 = alpha[2];
				a42 = alpha[3];
				a44 = alpha[4];
				filterM4();
				break;
			case 5:
				a11 = alpha[0];
				a31 = alpha[1];
				a33 = alpha[2];
				a51 = alpha[3];
				a53 = alpha[4];
				filterM5();
				break;
			default:
				break;
		}
	}
	
	
	public double[] getResponse() {
		return response;
	}
	
	public double[] getOrientation() {
		return orientation;
	}
	
	public void showResponse(String title) {
		IJtools.show(title, response, nx, ny);
	}	
	
	
	/**
	* Displays the orientation map in radiants (grayscale). 
	*/
	public void showOrientation(String title) {
		IJtools.show(title, orientation, nx, ny);
	}	

	
	/**
	* Displays the orientation map in HSB color, where the hue indicates orientation and brightness is proportional to the strength of the feature response. 
	*/
	public void showColorOrientation(String title) {
		byte[] h = new byte[size];
		byte[] s = new byte[size];
		byte[] b = new byte[size];
		double min = Double.MAX_VALUE;
		double max = -Double.MAX_VALUE;
		for (int i=0;i<size;i++) {
			if (response[i] > max) {
				max = response[i];
			}
			if (response[i] < min) {
				min = response[i];
			}
		}
		
		if (M % 2 != 0) {
			for (int i=0;i<size;i++) {
				h[i] = (byte)((orientation[i]/(2.0*PI) + 0.5)*255.0);
				s[i] = (byte)255;
				b[i] = (byte)((response[i]-min)/(max-min)*255.0);
			}
		} else {
			for (int i=0;i<size;i++) {
				h[i] = (byte)((orientation[i]/PI + 0.5)*255.0);
				s[i] = (byte)255;
				b[i] = (byte)((response[i]-min)/(max-min)*255.0);
			}
		}	
		ColorProcessor cp = new ColorProcessor(nx, ny);
		cp.setHSB(h, s, b);
		ImagePlus ip = new ImagePlus(title, cp);
		ip.show();
	}
	
	public int getWidth() {
		return nx;
	}
	
	public int getHeight() {
		return ny;
	}
	
	public String getOrder() {
		
		switch(M) {
			case 1:
				return "1st";
			case 2:
				return "2nd";
			case 3:
				return "3rd";
			case 4:
			case 5:
				return String.valueOf(M) + "th";
			default:
				return "";
		}
	}	

	
	/**
	* Computes and displays the response to rotations of the feature template over 2 pi, in (2*pi)/nIncrements increments.
	*/
	public void showRotations(int nIncrements) {
		IJtools.showStack(getOrder()+" order rotations", computeRotations(nIncrements), nx, ny, nIncrements);
	}
	
	
	/**
	* Performs non-maximum suppression on the filter response and displays the result. 
	*/
	public void showNMS() {
		IJtools.show("NMS", computeNMS(), nx, ny);
	}
	
		
	/**
	* Performs non-maximum suppression on the filter response. 
	*/
	public double[] computeNMS() {	
		
		double[] output = new double[size];
		
		double ux, uy, GA, GA1, GA2, angle;
		Interpolator interpolator;
		double[] slice = new double[nxy];
		int idx = 0;
		for (int z=0;z<nz;z++) {
			System.arraycopy(response, z*nxy, slice, 0, nxy);
			interpolator = new Interpolator(slice, nx, ny, "linear");
			for (int y=0;y<ny;y++) {		
				for (int x=0;x<nx;x++) {
					angle = orientation[idx];
					ux = -Math.sin(angle); //cos(angle + pi/2)
					uy = Math.cos(angle);
					GA1 = interpolator.getValue(x + ux, y + uy);
					GA2 = interpolator.getValue(x - ux, y - uy);
					GA = response[idx];
					if ( (GA < GA1) || (GA < GA2) ) {
						output[idx] = 0.0;
					} else {
						output[idx] = GA;
					}
					idx++;
				}
			}
		}
		return output;
	}
	
	/**
	* Computes and the response to rotations of the feature template over 2 pi, in (2*pi)/nIncrements increments.
	*/
	public double[] computeRotations(int nIncrements) {
		
		double[] rotations = new double[nxy*nIncrements];
		double radiantStep = 2.0*PI/(double)nIncrements;
		switch(M) {
			case 1:
				for (int k=0;k<nIncrements/2;k++) {
					for (int i=0;i<nxy;i++) {
						rotations[i+k*nxy] = pointRespM1(i, k*radiantStep);
						rotations[i+(k+nIncrements/2)*nxy] = -rotations[i+k*nxy];
					}
				}
				break;
			case 2:
				for (int k=0;k<nIncrements/2;k++) {
					for (int i=0;i<nxy;i++) {
						rotations[i+k*nxy] = pointRespM2(i, k*radiantStep);
						rotations[i+(k+nIncrements/2)*nxy] = rotations[i+k*nxy];
					}
				}
				break;
			case 3:
				for (int k=0;k<nIncrements/2;k++) {
					for (int i=0;i<nxy;i++) {
						rotations[i+k*nxy] = pointRespM3(i, k*radiantStep);
						rotations[i+(k+nIncrements/2)*nxy] = -rotations[i+k*nxy];
					}
				}
				break;
			case 4:
				for (int k=0;k<nIncrements/2;k++) {
					for (int i=0;i<nxy;i++) {
						rotations[i+k*nxy] = pointRespM4(i, k*radiantStep);
						rotations[i+(k+nIncrements/2)*nxy] = rotations[i+k*nxy];
					}
				}
				break;
			case 5:
				for (int k=0;k<nIncrements/2;k++) {
					for (int i=0;i<nxy;i++) {
						rotations[i+k*nxy] = pointRespM5(i, k*radiantStep);
						rotations[i+(k+nIncrements/2)*nxy] = -rotations[i+k*nxy];
					}
				}
				break;
			default:
				break;
		}
		return rotations;
	}
	
	
	private void filterM1() {
		double[] tRoots = new double[2];
		double gxi, gyi;
		double temp;
		
		//int iz = 0;
		double[] current = new double[nxy];
		
		for (int z=0;z<nz;z++) {
			System.arraycopy(input,z*nxy,current,0,nxy); // slice-by-slice processing
			computeBaseTemplates(current, nx, ny, M, sigma, z);
			int i = 0;
			for (int iz=0;iz<nxy && !stop;iz++) {
					gxi = gx[i];
					gyi = gy[i];
					gxi = approxZero(gxi, tolerance);
					gyi = approxZero(gyi, tolerance);
					
					if ( gxi==0.0 && gyi==0.0 ) {
						response[iz] = 0.0;
						orientation[iz] = 0.0;
					} else {   
						if (gyi == 0.0) {
							tRoots[0] = PI/2.0;
							tRoots[1] = opposite(tRoots[0]);
						} else {
							tRoots[0] = Math.atan(-gxi/gyi);
							tRoots[1] = opposite(tRoots[0]);
						}
						orientation[iz] = tRoots[0];
						response[iz] = Math.cos(tRoots[0])*a11*gyi - Math.sin(tRoots[0])*a11*gxi;				
						temp = Math.cos(tRoots[1])*a11*gyi - Math.sin(tRoots[1])*a11*gxi;
						if (temp > response[iz]) {
							response[iz] = temp;
							orientation[iz] = tRoots[1];
						}
					}
					i++;
				//}
			}
		}	
	}
	
	private void filterM2xx() { // nonseparable implementation, linear solution
		
		double d, gxyi;
		double temp;
		double[] tRoots = new double[2];
		
		int iz = 0;
		double[] current = new double[nxy];
		for (int z=0;z<nz;z++) {
			System.arraycopy(input,z*nxy,current,0,nxy);
			
			computeBaseTemplates(current, nx, ny, M, sigma, z);
			for (int i=0;i<nxy && !stop;i++) {
				
				d = approxZero(gxx[i] - gyy[i], tolerance);
				gxyi = approxZero(gxy[i], tolerance);
								
				if (d == 0.0 && gxyi == 0.0) {
					response[iz] = pointRespM2(i, 0.0);
					orientation[iz] = 0.0;
				} else {
					if (gxyi == 0.0) { // sin(2t) = 0, cos(2t) = 1 -> 2t = pi/2
						tRoots[0] = 0.0;
						tRoots[1] = -PI/2.0;
					} else {
						tRoots[0] = Math.atan(2.0*gxyi/d)/2.0;
						tRoots[1] = complement(tRoots[0]); //tRoots[0] + PI/2.0;
					}
					orientation[iz] = tRoots[0];
					response[iz] = pointRespM2(i, tRoots[0]);
					temp = pointRespM2(i, tRoots[1]);
					if (temp > response[iz]) {
						response[iz] = temp;
						orientation[iz] = tRoots[1];
					}
				}
				iz++;
			}
		}
	}
	
	private void filterM3() {
		
		double A, B, C, D;
		double temp;
		
		double a1 = 2.0*a31-3.0*a33;
		double a2 = 6.0*a33-7.0*a31;
		double g1, g2;
		
		int iz = 0;
		double[] current = new double[nxy];
		for (int z=0;z<nz;z++) {
			System.arraycopy(input,z*nxy,current,0,nxy);
			computeBaseTemplates(current, nx, ny, M, sigma, z);
			for (int i=0;i<nxy && !stop;i++) {
						
				g1 = -a11*gy[i];
				g2 = -a11*gx[i];
				
				A = g1 - a31*gyyy[i] + a1*gxxy[i];
				B = g2 + a2*gxyy[i] + a1*gxxx[i];
				C = g1 + a2*gxxy[i] + a1*gyyy[i];
				D = g2 - a31*gxxx[i] + a1*gxyy[i];
				
				A = approxZero(A, tolerance);
				B = approxZero(B, tolerance);
				C = approxZero(C, tolerance);
				D = approxZero(D, tolerance);

				if (A == 0.0) { // -> quadratic
					if (B == 0.0) { // -> linear
						if (C == 0.0) { // -> null, solve
							orientation[iz] = 0.0;
							response[iz] = pointRespM3(i, 0.0);
						} else { // solve linear
							if (D == 0.0) { // cos or sin can be 0
								double[] tRoots = new double[4];
								tRoots[0] = -PI/2.0;
								tRoots[1] = 0.0;
								tRoots[2] = PI/2.0;
								tRoots[3] = PI;
								response[iz] = pointRespM3(i, tRoots[0]);
								orientation[iz] = tRoots[0];
								for (int k=1;k<4;k++) {
									temp = pointRespM3(i, tRoots[k]);
									if (temp > response[iz]) {
										response[iz] = temp;
										orientation[iz] = tRoots[k];
									}
								}
							} else {
								double[] tRoots = new double[2];
								tRoots[0] = Math.atan(-D/C);
								tRoots[1] = opposite(tRoots[0]);
								response[iz] = pointRespM3(i, tRoots[0]);
								orientation[iz] = tRoots[0];
								temp = pointRespM3(i, tRoots[1]);
								if (temp > response[iz]) {
									response[iz] = temp;
									orientation[iz] = tRoots[1];
								}
							}	
						}
					} else { // solve quadratic
						double[] xRoots = IJmath.quadraticRoots(C/B, D/B);	
						double[] tRoots = new double[4];
						if (xRoots.length == 0 || xRoots[0] == 0.0 || xRoots[0] == -xRoots[1]) {
							tRoots[0] = -PI/2.0;
							tRoots[1] = 0.0;
							tRoots[2] = PI/2.0;
							tRoots[3] = PI;      			    	
						} else {  
							tRoots[0] = Math.atan(xRoots[0]);
							tRoots[1] = Math.atan(xRoots[1]);
							tRoots[2] = opposite(tRoots[0]);
							tRoots[3] = opposite(tRoots[1]);  
						} 
						response[iz] = pointRespM3(i, tRoots[0]);
						orientation[iz] = tRoots[0];
						for (int k=1;k<4;k++) {
							temp = pointRespM3(i, tRoots[k]);
							if (temp > response[iz]) {
								response[iz] = temp;
								orientation[iz] = tRoots[k];
							}
						}
					}	
				} else { // solve cubic
					double[] xRoots = IJmath.cubicRoots(B/A, C/A, D/A);
					double[] tRoots = new double[2*xRoots.length];
					double[] currentResponse = new double[tRoots.length];
					
					for (int k=0;k<xRoots.length;k++) {
						tRoots[k] = Math.atan(xRoots[k]);
						tRoots[k+xRoots.length] = opposite(tRoots[k]);
					}
					
					for (int k=0;k<tRoots.length;k++) {
						currentResponse[k] = pointRespM3(i, tRoots[k]);
					}
					sort(currentResponse, tRoots);
					
					if (xRoots.length == 3) {
					if (approxEqual(currentResponse[tRoots.length-1], currentResponse[tRoots.length-2], 0.000001)) {
						response[iz] = currentResponse[tRoots.length-3];
						orientation[iz] = tRoots[tRoots.length-3];
					} else {
						response[iz] = currentResponse[tRoots.length-1];
						orientation[iz] = tRoots[tRoots.length-1];
					}
					} else {
						response[iz] = currentResponse[tRoots.length-1];
						orientation[iz] = tRoots[tRoots.length-1];
					
					}		
				}
				iz++;
			}
			
		}
	}	
	

	
	private void filterM4() {
		
		double g1, g2, g3;
		double A, B, C, D, E;
		double[] xRoots, tRoots;
		double temp;
		int N;
		double tolerance = 0.000000000001;
		
		double a1 = 2.0*a44-a42;
		double a2 = 2.0*a40-a42;
		double a3 = a22-a20;
		double a4 = 6.0*(a44-a42+a40);
		int iz = 0;
		double[] current = new double[nxy];
		for (int z=0;z<nz;z++) {
			System.arraycopy(input,z*nxy,current,0,nxy);
			computeBaseTemplates(current, nx, ny, M, sigma, z);
			for (int i=0;i<nxy && !stop;i++) {
					
				g1 = a3*gxy[i];
				g2 = a3*(gxx[i]-gyy[i]);
				g3 = a4*gxxyy[i];

				A = a1*gxxxy[i] - a2*gxyyy[i] + g1;
				B = a1*gxxxx[i] + a2*gyyyy[i] - g3 + g2;
				C = a4*(gxyyy[i] - gxxxy[i]);
				D = -a1*gyyyy[i] - a2*gxxxx[i] + g3 + g2;
				E = a2*gxxxy[i] - a1*gxyyy[i] - g1;
				
				A = approxZero(A, tolerance);
				C = approxZero(C, tolerance);
				E = approxZero(E, tolerance);
				

				if (A == 0.0) { // -> cubic
					if (B == 0.0) { // -> quadratic
						if (C == 0.0) { // -> linear
							if (D == 0.0) { // solve null
								response[iz] = pointRespM4(i, 0.0);
								orientation[iz] = 0.0;
							} else { // solve linear
								orientation[iz] = Math.atan(-E/D);
								response[iz] = pointRespM4(i, orientation[iz]);
							}
						} else { // solve quadratic
							xRoots = IJmath.quadraticRoots(D/C, E/C);
							N = xRoots.length;
							if (N == 0) {
								response[iz] = pointRespM4(i, 0.0);
								orientation[iz] = 0.0;					
							} else {
								tRoots = new double[2];
								tRoots[0] = Math.atan(xRoots[0]);
								tRoots[1] = Math.atan(xRoots[1]);
								response[iz] = pointRespM4(i, tRoots[0]);
								orientation[iz] = tRoots[0];
								temp = pointRespM4(i, tRoots[1]);
								if (temp > response[iz]) {
									response[iz] = temp;
									orientation[iz] = tRoots[1];
								}
							}
						}			
					}	else { // solve cubic -> (B != 0)
						if ( (C == 0.0) && (E == 0.0) ) {
							double deltaQ = -D/B;
							if (deltaQ >= 0.0) { // Bx^3 + Dx = 0						
								tRoots = new double[4];
								tRoots[0] = Math.atan(Math.sqrt(deltaQ));
								tRoots[1] = Math.atan(-Math.sqrt(deltaQ));
								tRoots[2] = 0.0;
								tRoots[3] = PI/2.0;

								response[iz] = pointRespM4(i, tRoots[0]);
								orientation[iz] = tRoots[0];
								for (int k=1;k<4;k++) {
									temp = pointRespM4(i, tRoots[k]);
									if (temp > response[iz]) {
										response[iz] = temp;
										orientation[iz] = tRoots[k];
									}	
								}
							}	else {
								tRoots = new double[2];
								tRoots[0] = 0.0;
								tRoots[1] = PI/2.0;
								
								response[iz] = pointRespM4(i, 0.0);
								orientation[iz] = 0.0;
								temp = pointRespM4(i, PI/2.0);
								if (temp > response[iz]) {
									response[iz] = temp;
									orientation[iz] = PI/2.0;
								}
							}
						} else {
							xRoots = IJmath.cubicRoots(C/B, D/B, E/B);
							N = xRoots.length;
							tRoots = new double[N];
							for (int k=0;k<N;k++) {
								tRoots[k] = Math.atan(xRoots[k]);
							}
							response[iz] = pointRespM4(i, tRoots[0]);
							orientation[iz] = tRoots[0];
							for (int k=1;k<N;k++) {
								temp = pointRespM4(i, tRoots[k]);
								if (temp > response[iz]) {
									response[iz] = temp;
									orientation[iz] = tRoots[k];
								}	
							}
						}
					}	
				} else { // solve quartic		
					
					xRoots = IJmath.quarticRoots(B/A, C/A, D/A, E/A);
					N = xRoots.length;
					if (N == 0) {
						orientation[iz] = 0.0;
						response[iz] = pointRespM4(i, 0.0);
					} else {
						tRoots = new double[N];
						for (int k=0;k<N;k++) {
							tRoots[k] = Math.atan(xRoots[k]);
						}
						response[iz] = pointRespM4(i, tRoots[0]);
						orientation[iz] = tRoots[0];
						for (int k=1;k<N;k++) {
							temp = pointRespM4(i, tRoots[k]);
							if (temp > response[iz]) {
								response[iz] = temp;
								orientation[iz] = tRoots[k];
							}	
						}
					}
				}
				iz++;
			}
		}
	}
	

	private void filterM5() {
		//ProgressBarManager pb = new ProgressBarManager(nx);
		filterM1();
		double[] initialAngle = getOrientation();
		double a1 = 2.0*a31 - 3.0*a33;
		double a2 = 4.0*a51 - 3.0*a53;
		double a3 = 6.0*a33 - 7.0*a31;
		double a4 = 12.0*a51 - 17.0*a53;
		double a5 = 6.0*a53 - 13.0*a51;
		double a6 = 3.0*a33 - 5.0*a31;
		double a7 = a31 - 3.0*a33;
		double a8 = 30.0*a53 - 34.0*a51;
		double a53x2 = 2.0*a53;
		double a11x2 = 2.0*a11;
		double[] quinticC = new double[6];
		double[] quarticC = new double[5];
		double[] cubicC = new double[4];
		double[] quarticRoots;
		double[] roots, tRoots;
		int N;
		double x0;
		double[] f0;
		double[] croot;
		double theta;
		int iz = 0;
		double[] current = new double[nxy];
		//double[] map = new double[nxy];
		
		
		for (int z=0;z<nz;z++) {
			System.arraycopy(input,z*nxy,current,0,nxy);
			computeBaseTemplates(current, nx, ny, M, sigma, z);
			for (int i=0;i<nxy && !stop;i++) {
				//IJ.showProgress(i,nxy-1);
				quinticC[0] = -a11*gx[i] + a1*gxyy[i] - a31*gxxx[i] - a51*gxxxxx[i]  + a2*gxxxyy[i] + a53x2*gxyyyy[i];
				quinticC[1] = -a11*gy[i] + a3*gxxy[i] + a5*gxxxxy[i] + a1*gyyy[i] + a4*gxxyyy[i] + a53x2*gyyyyy[i];
				quinticC[2] = -a11x2*gx[i] + a7*gxxx[i] + a2*gxxxxx[i] + a6*gxyy[i] + a8*gxxxyy[i] + a4*gxyyyy[i];
				quinticC[3] = -a11x2*gy[i] + a6*gxxy[i] + a4*gxxxxy[i] + a7*gyyy[i] + a8*gxxyyy[i] + a2*gyyyyy[i];
				quinticC[4] = -a11*gx[i] + a1*gxxx[i] + a53x2*gxxxxx[i] + a3*gxyy[i] + a4*gxxxyy[i] + a5*gxyyyy[i];
				quinticC[5] = -a11*gy[i] + a1*gxxy[i] - a31*gyyy[i] + a53x2*gxxxxy[i] + a2*gxxyyy[i] - a51*gyyyyy[i];
				
				quinticC[5] = approxZero(quinticC[5], tolerance);
				quinticC[4] = approxZero(quinticC[4], tolerance);
				quinticC[3] = approxZero(quinticC[3], tolerance);
				quinticC[2] = approxZero(quinticC[2], tolerance);
				quinticC[1] = approxZero(quinticC[1], tolerance);
				quinticC[0] = approxZero(quinticC[0], tolerance); // check order!! -> highest degree coeff -> [5]
				
				if (quinticC[5] == 0.0) {
					if (quinticC[4] == 0.0) {
						if (quinticC[3] == 0.0) {
							if (quinticC[2] == 0.0) {
								if (quinticC[1] == 0.0) {
									roots = new double[1];
									roots[0] = 0.0;
								} else {
									roots = new double[1];
									roots[0] = -quinticC[0]/quinticC[1];
								}
							} else {	
								roots = IJmath.quadraticRoots(quinticC[1]/quinticC[2], quinticC[0]/quinticC[2]);
							}
						} else {
							roots = IJmath.cubicRoots(quinticC[2]/quinticC[3], quinticC[1]/quinticC[3], quinticC[0]/quinticC[3]);
						}
					} else {
						roots = IJmath.quarticRoots(quinticC[3]/quinticC[4], quinticC[2]/quinticC[4], quinticC[1]/quinticC[4], quinticC[0]/quinticC[4]);
					}

					N = roots.length;
					
					if (N != 0) {
						tRoots = new double[2*N+4];
						for (int k=0;k<N;k++) {
							tRoots[k] = Math.atan(roots[k]);
							tRoots[k+N] = opposite(tRoots[k]);
						}
						tRoots[2*N] = 0.0;
						tRoots[2*N+1] = PI/2.0;
						tRoots[2*N+2] = PI;
						tRoots[2*N+3] = -PI/2.0;

						response[iz] = pointRespM5(i, tRoots[0]);
						orientation[iz] = tRoots[0];
						for (int k=1;k<tRoots.length;k++) {
							double temp = pointRespM5(i, tRoots[k]);
							if (temp > response[iz]) {
								response[iz] = temp;
								orientation[iz] = tRoots[k];
							}	
						}
					} else {
						tRoots = new double[4];//{-PI/2.0, 0, PI/2.0, PI};
						tRoots[0] = 0.0;
						tRoots[1] = PI/2.0;
						tRoots[2] = PI;
						tRoots[3] = -PI/2.0;
						response[iz] = pointRespM5(i, tRoots[0]);
						orientation[iz] = tRoots[0];
						for (int k=1;k<tRoots.length;k++) {
							double temp = pointRespM5(i, tRoots[k]);
							if (temp > response[iz]) {
								response[iz] = temp;
								orientation[iz] = tRoots[k];
							}	
						}						
					}

				} else {
					theta = initialAngle[iz];
					x0 = Math.tan(theta);
					f0 = IJmath.evalPolyD(quinticC, x0);
					
					if (quinticC[0] == 0.0) { // root = 0.0
						quarticRoots = IJmath.quarticRoots(quinticC[4]/quinticC[5], quinticC[3]/quinticC[5], quinticC[2]/quinticC[5], quinticC[1]/quinticC[5]);
						N = quarticRoots.length;
						roots = new double[N+1];
						roots[0] = 0.0;
						for (int k=1;k<=N;k++) {
							roots[k] = quarticRoots[k-1];
						}
						N++;
					} else if (f0[0] == 0.0) { // root = x0
						quarticC = IJmath.divPolyByRoot(quinticC, x0);
						quarticRoots = IJmath.quarticRoots(quarticC[3]/quarticC[4], quarticC[2]/quarticC[4], quarticC[1]/quarticC[4], quarticC[0]/quarticC[4]);
						N = quarticRoots.length;
						roots = new double[N+1];
						roots[0] = x0;
						for (int k=1;k<=N;k++) {
							roots[k] = quarticRoots[k-1];
						}
						N++;
					} else {
						if (Math.abs(x0) >= 100.0) {
							x0 = 0.0;
						}
						croot = IJmath.laguerre(quinticC, x0);
						//map[i] = converged[0];
						croot[0] = approxZero(croot[0], 1e-15);
						croot[1] = approxZero(croot[1], 1e-15);

						if (croot[1] == 0) { // real root
							quarticC = IJmath.divPolyByRoot(quinticC, croot[0]);
							quarticRoots = IJmath.quarticRoots(quarticC[3]/quarticC[4], quarticC[2]/quarticC[4], quarticC[1]/quarticC[4], quarticC[0]/quarticC[4]);
							N = quarticRoots.length;
							roots = new double[N+1];
							roots[0] = croot[0];
							for (int k=1;k<=N;k++) {
								roots[k] = quarticRoots[k-1];
							}
							N++;
						} else { // complex root
							cubicC = IJmath.divPolyByConjRoots(quinticC, croot[0], croot[1]);
							roots = IJmath.cubicRoots(cubicC[2]/cubicC[3], cubicC[1]/cubicC[3], cubicC[0]/cubicC[3]);
							N = roots.length;
						}
						
					}
					
					tRoots = new double[2*N];
					for (int k=0;k<N;k++) {
						tRoots[k] = Math.atan(roots[k]);
						tRoots[k+N] = opposite(tRoots[k]);
					}
					
					response[iz] = pointRespM5(i, tRoots[0]);
					orientation[iz] = tRoots[0];
					for (int k=1;k<tRoots.length;k++) {
						double temp = pointRespM5(i, tRoots[k]);
						if (temp > response[iz]) {
							response[iz] = temp;
							orientation[iz] = tRoots[k];
						}	
					}
				}
				iz++;
			}
		}
	}
	
	static private void sort(double[] responses, double[] angles) {
	
		int N = responses.length;
		double temp;
		double min;
		int minIndex;
		for (int i=0;i<N-1;i++) {
			min = responses[i];
			minIndex = i;
			for (int k=i+1;k<N;k++) {
				if (responses[k] < min) {
					min = responses[k];
					minIndex = k;
				}
			}
			temp = responses[i];
			responses[i] = min;
			responses[minIndex] = temp;
			temp = angles[i];
			angles[i] = angles[minIndex];
			angles[minIndex] = temp;
		}
	}	
	
	private double opposite(double theta) {
	    if (theta > 0.0) { // values in (-PI,PI]
	        return theta - PI;
	    } else {
	        return theta + PI;
	    }	
	}
	
	
	private double complement(double theta) { // complementary angle from [-PI/4, PI/4( in [-PI/2, PI/2(
	    if (theta > 0.0) {
	        return theta - PI/2.0;
	    } else {
	        return theta + PI/2.0;
	    }	
	}
	
	
	private double approxZero(double n, double tol) {
		if (Math.abs(n) < tol) {
			return 0.0;
		} else {
		    return n;
		}    
	}
	
	private boolean approxEqual(double a, double b, double tol) {
		
		if ( Math.abs(1-a/b) < tol) {
			return true;
		} else {
			return false;
		}		 
	}
	
	
	private void computeBaseTemplates(double[] input, int nx, int ny, int M, double sigma, int z) {
		int wWidth = (int)(4.0*sigma);
		int kLength = wWidth+1;

		double[] aKernel = new double[kLength];
		double[] bKernel = new double[kLength];
		double g[] = new double[kLength];
		double d;
		double sigma2 = sigma*sigma;
		double sigma4 = sigma2*sigma2;
		double sigma6 = sigma4*sigma2;
		double sigma8 = sigma4*sigma4;
		double sigma10 = sigma6*sigma4;
		double sigma12 = sigma8*sigma4;
		
		for (int i=0;i<kLength;i++) {
			g[i] = Math.exp(-(i*i)/(2.0*sigma2));
		}
		
		if (M == 1 || M == 3 || M == 5) {
			
			d = 2.0*PI*sigma4;
			for (int i=0;i<kLength;i++) {
				aKernel[i] = -i*g[i] / d;
			}
						
			gx = Convolver2D.convolveOddX(input, aKernel, nx, ny);
			gx = Convolver2D.convolveEvenY(gx, g, nx, ny);
			gy = Convolver2D.convolveOddY(input, aKernel, nx, ny);
			gy = Convolver2D.convolveEvenX(gy, g, nx, ny);


			if (M == 3 || M == 5) {
				
				d = 2.0*PI*sigma8;
				for (int i=0;i<kLength;i++) {
					aKernel[i] = (3.0*i*sigma2 - i*i*i) * g[i] / d;
				}
				gxxx = Convolver2D.convolveOddX(input, aKernel, nx, ny);
				gxxx = Convolver2D.convolveEvenY(gxxx, g, nx, ny);
				gyyy = Convolver2D.convolveOddY(input, aKernel, nx, ny);
				gyyy = Convolver2D.convolveEvenX(gyyy, g, nx, ny);
				for (int i=0;i<kLength;i++) {
					aKernel[i] = (sigma2 - i*i) * g[i] / d;
					bKernel[i] = i*g[i];
				}
				gxxy = Convolver2D.convolveEvenX(input, aKernel, nx, ny);
				gxxy = Convolver2D.convolveOddY(gxxy, bKernel, nx, ny);
				gxyy = Convolver2D.convolveEvenY(input, aKernel, nx, ny);
				gxyy = Convolver2D.convolveOddX(gxyy, bKernel, nx, ny);
			}
			if (M == 5) {
				
				d = 2.0*PI*sigma12;
				for (int i=0;i<kLength;i++) {
					aKernel[i] = -i*(i*i*i*i - 10.0*i*i*sigma2 + 15.0*sigma4) * g[i] / d;
				}
				gxxxxx = Convolver2D.convolveOddX(input, aKernel, nx, ny);
				gxxxxx = Convolver2D.convolveEvenY(gxxxxx, g, nx, ny);
				gyyyyy = Convolver2D.convolveOddY(input, aKernel, nx, ny);
				gyyyyy = Convolver2D.convolveEvenX(gyyyyy, g, nx, ny);
				for (int i=0;i<kLength;i++) {
					aKernel[i] = (i*i*i*i - 6.0*i*i*sigma2 + 3.0*sigma4) * g[i] / d;
					bKernel[i] = -i * g[i];
				}
				gxxxxy = Convolver2D.convolveEvenX(input, aKernel, nx, ny);
				gxxxxy = Convolver2D.convolveOddY(gxxxxy, bKernel, nx, ny);
				gxyyyy = Convolver2D.convolveEvenY(input, aKernel, nx, ny);
				gxyyyy = Convolver2D.convolveOddX(gxyyyy, bKernel, nx, ny);
				for (int i=0;i<kLength;i++) {
					aKernel[i] = i*(i*i - 3.0*sigma2) * g[i] / d;
					bKernel[i] = (sigma2 - i*i) * g[i];
				}
				gxxxyy = Convolver2D.convolveOddX(input, aKernel, nx, ny);
				gxxxyy = Convolver2D.convolveEvenY(gxxxyy, bKernel, nx, ny);
				gxxyyy = Convolver2D.convolveOddY(input, aKernel, nx, ny);
				gxxyyy = Convolver2D.convolveEvenX(gxxyyy, bKernel, nx, ny);		
			}
		} else { //(M == 2 || M == 4)
		
			d = 2.0*PI*sigma6;
			for (int i=0;i<kLength;i++) {
				aKernel[i] = (i*i - sigma2) * g[i] / d;
			}
			gxx = Convolver2D.convolveEvenX(input, aKernel, nx, ny);
			gxx = Convolver2D.convolveEvenY(gxx, g, nx, ny);
			gyy = Convolver2D.convolveEvenY(input, aKernel, nx, ny);
			gyy = Convolver2D.convolveEvenX(gyy, g, nx, ny);
			for (int i=0;i<kLength;i++) {
				aKernel[i] = i * g[i];
				bKernel[i] = aKernel[i] / d;
			}
			gxy = Convolver2D.convolveOddX(input, aKernel, nx, ny);
			gxy = Convolver2D.convolveOddY(gxy, bKernel, nx, ny);
			
			if (M == 4) {
			
				d = 2.0*PI*sigma10;
				for (int i=0;i<kLength;i++) {
					aKernel[i] = (i*i*i*i - 6.0*i*i*sigma2 + 3.0*sigma4) * g[i] / d;
				}
				gxxxx = Convolver2D.convolveEvenX(input, aKernel, nx, ny);
				gxxxx = Convolver2D.convolveEvenY(gxxxx, g, nx, ny);
				gyyyy = Convolver2D.convolveEvenY(input, aKernel, nx, ny);
				gyyyy = Convolver2D.convolveEvenX(gyyyy, g, nx, ny);
				for (int i=0;i<kLength;i++) {
					aKernel[i] = i * (i*i - 3.0*sigma2) * g[i] / d;
					bKernel[i] = i * g[i];
				}
				gxxxy = Convolver2D.convolveOddX(input, aKernel, nx, ny);
				gxxxy = Convolver2D.convolveOddY(gxxxy, bKernel, nx, ny);
				gxyyy = Convolver2D.convolveOddY(input, aKernel, nx, ny);
				gxyyy = Convolver2D.convolveOddX(gxyyy, bKernel, nx, ny);
				for (int i=0;i<kLength;i++) {	
					aKernel[i] = (sigma2 - i*i) * g[i];
					bKernel[i] = aKernel[i] / d;
				}
				gxxyy = Convolver2D.convolveEvenX(input, aKernel, nx, ny);
				gxxyy = Convolver2D.convolveEvenY(gxxyy, bKernel, nx, ny);
			}
		}
	}
	
	
	
	
	
	// Point responses
	private double pointRespM1(int i, double angle) {
		
		double cosT = Math.cos(angle);
		double sinT = Math.sin(angle);
		
		double result = a11 * (cosT*gy[i] - sinT*gx[i]);
		return result;
	}
	
	
	private double pointRespM2(int i, double angle) {
		
		double cosT = Math.cos(angle);
		double sinT = Math.sin(angle);
		
		double result = cosT*cosT*(a22*gyy[i] + a20*gxx[i])
			   		  + cosT*sinT*2.0*(a20-a22)*gxy[i]
			   		  + sinT*sinT*(a22*gxx[i] + a20*gyy[i]);
		return result;
	}
	
	
	private double pointRespM3(int i, double angle) {
	
		double cosT = Math.cos(angle);
		double sinT = Math.sin(angle);
		double cosT2 = cosT*cosT;
		double sinT2 = sinT*sinT;
		
		double result = cosT*a11*gy[i]
					  - sinT*a11*gx[i]
					  + cosT2*cosT * (a31*gxxy[i] + a33*gyyy[i])
					  - sinT2*sinT * (a31*gxyy[i] + a33*gxxx[i])
					  - cosT2*sinT * (3.0*a33*gxyy[i] - 2.0*a31*gxyy[i] + a31*gxxx[i] )
					  + cosT*sinT2 * (3.0*a33*gxxy[i] - 2.0*a31*gxxy[i] + a31*gyyy[i] );
		return result;
	}
	
	
	private double pointRespM4(int i, double angle) {
	
		double cosT = Math.cos(angle);
		double sinT = Math.sin(angle);
		double cosTsinT = cosT*sinT;
		double cosT2 = cosT*cosT;
		double sinT2 = sinT*sinT;
		double a = (a20-a22)*gxy[i];

		double result = cosT2*cosT2 * (a20*gxx[i]+a22*gyy[i]+a40*gxxxx[i]+a42*gxxyy[i]+a44*gyyyy[i])
					  + cosT2*cosTsinT * 2.0*(a+2.0*a40*gxxxy[i]+a42*(gxyyy[i]-gxxxy[i])-2.0*a44*gxyyy[i])
					  + cosT2*sinT2 * ((a20+a22)*(gxx[i]+gyy[i])+a42*(gxxxx[i]+gyyyy[i])+(6.0*(a40+a44)-4.0*a42)*gxxyy[i])
					  + sinT2*cosTsinT * 2.0*(a+2.0*a40*gxyyy[i]+a42*(gxxxy[i]-gxyyy[i])-2.0*a44*gxxxy[i])
					  + sinT2*sinT2 * (a20*gyy[i]+a22*gxx[i]+a40*gyyyy[i]+a42*gxxyy[i]+a44*gxxxx[i]);
		return result;
	}
	
	
	private double pointRespM5(int i, double angle) {
	
		double cosT = Math.cos(angle);
		double sinT = Math.sin(angle);
		double cosT2 = cosT*cosT;
		double sinT2 = sinT*sinT;
		double cosT3 = cosT2*cosT;
		double sinT3 = sinT2*sinT;
			
		double result = cosT2*cosT3 * (a11*gy[i] + a31*gxxy[i] + a33*gyyy[i] + a51*gxxxxy[i] + a53*gxxyyy[i])
					  + cosT2*cosT2*sinT * (-a11*gx[i] - a31*gxxx[i] - (3.0*a33-2.0*a31)*gxyy[i] + a51*(4.0*gxxxyy[i] - gxxxxx[i]) + a53*(2.0*gxyyyy[i] - 3.0*gxxxyy[i]))
					  + cosT3*sinT2 * ( 2.0*a11*gy[i] + (3.0*a33-a31)*gxxy[i] + (a33+a31)*gyyy[i] + a51*(6.0*gxxyyy[i]-4.0*gxxxxy[i]) + a53*(gyyyyy[i]-6.0*gxxyyy[i]+3.0*gxxxxy[i]))
					  + cosT2*sinT3 * (-2.0*a11*gx[i] - (3.0*a33-a31)*gxyy[i] - (a33+a31)*gxxx[i] - a51*(6.0*gxxxyy[i]-4.0*gxyyyy[i]) - a53*(gxxxxx[i]-6.0*gxxxyy[i]+3.0*gxyyyy[i]))
					  + cosT*sinT2*sinT2 * ( a11*gy[i] + a31*gyyy[i] + (3.0*a33-2.0*a31)*gxxy[i] - a51*(4.0*gxxyyy[i] - gyyyyy[i]) - a53*(2.0*gxxxxy[i] - 3.0*gxxyyy[i]))
					  - sinT2*sinT3 * (a11*gx[i] + a31*gxyy[i] + a33*gxxx[i] + a51*gxyyyy[i] + a53*gxxxyy[i]);
		return result;
	}

	
		
}