/*====================================================================
| Version: December 7, 2010
\===================================================================*/

/*====================================================================
| Ricard Delgado-Gonzalo
| EPFL/STI/IOA/LIB/BM.4.141
| e-mail: ricard.delgado@gmail.com
| URL: http://www.ricarddelgado.com/
\===================================================================*/
package segmentation;

import java.awt.Polygon;
import java.awt.Rectangle;
import java.io.IOException;
import java.util.Arrays;

import javax.swing.JProgressBar;
import javax.swing.JTextArea;


import analysis.CTCellAnalyzer;
import big.ij.snake2D.Snake2DKeeper;
import big.ij.snake2D.Snake2DScale;

import filters.CTBinaryDilation;
import filters.CTMinMax;

import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.gui.Roi;
import ij.plugin.filter.GaussianBlur;
import ij.process.FloatProcessor;
import ij.process.ImageProcessor;
import io.CTInput;
import structures.CTExtractedData;
import structures.CTRunParameters;
import structures.CTUnsortedPairSortedList;

public class CTSegmentation {
	private int Nstack;
	private String inputFolderCh1 			= null;
	private String inputFolderCh2 			= null;
	private String[] fileListCh1 			= null;
	private String[] fileListCh2 			= null;
	private CTRunParameters runParameters 	= null;
	private CTExtractedData[] extractedData = null;
	private JProgressBar progressBar 		= null;
	private JTextArea taskLog 				= null;

	private static int LAPLACIAN_WIDTH 	= 3;
    private static int LAPLACIAN_HEIGHT = 3;
	private static float LAPLACIAN_KERNEL [] = {(float)0.0,(float)-1.0,(float)0.0, (float)-1.0,(float)4.0,(float)-1.0, (float)0.0,(float)-1.0,(float)0.0}; 

	private static double GAUSSIANACCURACY = 0.000001;
	
	private CTWatershedProcessing watershedProcessing;
	private GaussianBlur gaussianBlur;
	private CTMinMax minmaxFilter;

	private boolean[] regionMask = null;
	private int minVal, maxVal;
	private boolean successful = true;
	private boolean showWatershedRegions = false;
	private boolean showTempStd = false;
	private boolean showNMRMap = false;
	
	public CTSegmentation(String inputFolderCh1, String[] fileListCh1, String inputFolderCh2, String[] fileListCh2, CTRunParameters runParameters, CTExtractedData[] extractedData){
		this.inputFolderCh1 = inputFolderCh1;
		this.inputFolderCh2 = inputFolderCh2;
		this.fileListCh1 = fileListCh1;
		this.fileListCh2 = fileListCh2;
		this.runParameters = runParameters;
		this.extractedData = extractedData;
		Nstack = extractedData.length;
		gaussianBlur = new GaussianBlur();
		minmaxFilter = new CTMinMax();
	}
	public void setProgressBar(JProgressBar progressBar){
		this.progressBar = progressBar;
	}
	
	public void showWatershedRegions(boolean showWatershedRegions){
		this.showWatershedRegions = showWatershedRegions;
	}

	public void showTempStd(boolean showTempStd){
		this.showTempStd = showTempStd;
	}

	public void showNMRMap(boolean showNMRMap){
		this.showNMRMap = showNMRMap;
	}

	
	public void setLog(JTextArea taskLog){
		this.taskLog = taskLog;
	}
	
	private void writeLog(String str){
		if(taskLog!=null){
			taskLog.append(str+"\n");
			taskLog.setCaretPosition(taskLog.getDocument().getLength());
		}else{
			IJ.log(str);
		}
	}
	
	public void doSegmentation(){
		if(progressBar!=null) progressBar.setValue(0);
		Roi roi = new Roi(runParameters.roiBox);
		computeRegionMask(roi);
		
		for(int frame=0; frame<Nstack && successful; frame++) {
			writeLog("Analyzing frame "+ (frame+1) + " of " + Nstack);
			FloatProcessor ipr1 = CTInput.readProcessor(inputFolderCh1,fileListCh1[frame], roi);
			FloatProcessor ipr2 = CTInput.readProcessor(inputFolderCh2,fileListCh2[frame], roi);
			successful = watershedSegmentation(frame, ipr1, ipr2);
			if(progressBar!=null) progressBar.setValue(frame);
		}
		
		if(showWatershedRegions && successful){
			writeLog("Showing watershed regions");
			ImageStack watershedRegionsStack = new ImageStack(runParameters.roiBox.width, runParameters.roiBox.height);
			for(int i=0; i<Nstack; i++){
				watershedRegionsStack.addSlice("frame "+i,extractedData[i].getRegionLabelMap());
			}
			ImagePlus watershedRegionsImagPlus = new ImagePlus("Watershed regions",watershedRegionsStack);
			watershedRegionsImagPlus.show();
		}
	}
	
	private boolean watershedSegmentation(int frame, FloatProcessor ipr1, FloatProcessor ipr2){
		ImageProcessor tmp_imProcessor = ipr1.duplicate();
		minmaxFilter.filter2DMINMAX(tmp_imProcessor, CTMinMax.MAX, 1, 1);
		gaussianBlur.blurFloat((FloatProcessor)tmp_imProcessor, runParameters.sigma, runParameters.sigma, GAUSSIANACCURACY);
		watershedProcessing = new CTWatershedProcessing(tmp_imProcessor, regionMask, minVal, maxVal, runParameters.regionOffset);
		watershedProcessing.setLog(taskLog);
		
		writeLog("Performing watershed...");
		boolean watershedSuccess = watershedProcessing.watershed();
		writeLog("Watershed finished");
		
		if(watershedSuccess){
			writeLog("Extracting data from image...");
			extractedData[frame] = new CTExtractedData();
			extractedData[frame].setRegionLabelMap(watershedProcessing.getLabelsMap());
			extractedData[frame].setNeighborhoodMatrix(watershedProcessing.getNeighborhoodMatrix());
			extractedData[frame].setRegionCentroidX(watershedProcessing.getCentroidsX());
			extractedData[frame].setRegionCentroidY(watershedProcessing.getCentroidsY());
			extractedData[frame].setRegionArea(watershedProcessing.getArea());
			extractedData[frame].setNregions(watershedProcessing.getNregions());
			extractedData[frame].setRegionXmin(watershedProcessing.getXminRegions());
			extractedData[frame].setRegionXmax(watershedProcessing.getXmaxRegions());
			extractedData[frame].setRegionYmin(watershedProcessing.getYminRegions());
			extractedData[frame].setRegionYmax(watershedProcessing.getYmaxRegions());
			extractedData[frame].extractRegionImageValuesCh1(ipr1);
			extractedData[frame].extractRegionImageValuesCh2(ipr2);
			analyzeCell(extractedData[frame]);
		}else{
			writeLog("No regions detected in the imaging area.");
			writeLog("Segmentation process aborted.");
		}
		writeLog("Extraction finished");		
		return watershedSuccess;
	}

	//TODO remove this function
	private double[] getParameters(){
		String path = null;
		double[] par = null;
		try {
			path =  new java.io.File(".").getCanonicalPath();
		} catch (IOException e) {
			e.printStackTrace();
			par = new double[2];
			par[0] = 0;
			par[1] = 0;
			return par;
		}
		
		par = CTInput.loadParamXML(path+System.getProperty("file.separator")+"param.xml");
		if(par==null){
			par = new double[2];
			par[0] = 0;
			par[1] = 0;
			return par;
		}
		return par;
	}
	
	private void analyzeCell(CTExtractedData cTExtractedData) {
		int nx = cTExtractedData.getRegionLabelMap().getWidth();
		int ny = cTExtractedData.getRegionLabelMap().getHeight();
		double[][] rawDataCh1 = cTExtractedData.getRegionImageValuesCh1();
		double[][] rawDataCh2 = cTExtractedData.getRegionImageValuesCh2();
		int Nregions = cTExtractedData.getNregions();
		FloatProcessor labelMap = cTExtractedData.getRegionLabelMap();
		float[] labelMapPixels = (float[])labelMap.getPixels();
		int[] regionWidth = cTExtractedData.getRegionWidth();
		int[] regionHeight = cTExtractedData.getRegionHeight();
		int[] regionXmin = cTExtractedData.getRegionXmin();
		int[] regionXmax = cTExtractedData.getRegionXmax();
		int[] regionYmin = cTExtractedData.getRegionYmin();
		int[] regionYmax = cTExtractedData.getRegionYmax();
		double[] regionArea = cTExtractedData.getRegionArea();
		
		double[] regionMeanCh1 							= new double[Nregions];
		double[] regionMeanCh2 							= new double[Nregions];
		double[] cellMeanCh1 							= new double[Nregions];
		double[] cellMeanCh2 							= new double[Nregions];
		double[] cellTop3MeanCh1						= new double[Nregions];
		double[] cellTop3MeanCh2						= new double[Nregions];
		double[] cellTop5MeanCh1						= new double[Nregions];
		double[] cellTop5MeanCh2						= new double[Nregions];
		double[] cellMedianCh1							= new double[Nregions];
		double[] cellMedianCh2							= new double[Nregions];
		Polygon[] cellOutline 							= new Polygon[Nregions];
		int[] segmentationType 							= new int[Nregions];
		double[] segmentationQuality 					= new double[Nregions];
		double[] cellCircularity 						= new double[Nregions];
		double[] cellArea 								= new double[Nregions];
		double[] minValueRegionCh1 						= new double[Nregions];
		double[] minValueRegionCh2 						= new double[Nregions];
		double[] minValueRegionNotCellCh1 				= new double[Nregions];
		double[] minValueRegionNotCellCh2 				= new double[Nregions];
		double[] regionBottomNPercentValueCh1 			= new double[Nregions];
		double[] regionBottomNPercentValueCh2 			= new double[Nregions];
		double[] regionNotCellBottomNPercentValueCh1	= new double[Nregions];
		double[] regionNotCellBottomNPercentValueCh2 	= new double[Nregions];
		double[] cellMeanVScellStdCh1 					= new double[Nregions];
		double[] cellMeanVScellStdCh2 					= new double[Nregions];
		double[] regionNotCellMeanVSCellMeanCh1 		= new double[Nregions];
		double[] regionNotCellMeanVSCellMeanCh2 		= new double[Nregions];
		double[] energyLoG10Ch1							= new double[Nregions];
		double[] energyLoG10Ch2							= new double[Nregions];
		double[] energyLoG3Ch1							= new double[Nregions];
		double[] energyLoG3Ch2							= new double[Nregions];
		double[] energySteerable4Ch1					= new double[Nregions];
		double[] energySteerable4Ch2					= new double[Nregions];
		
		double[] param = getParameters();
		
		for(int i = 0; i < Nregions; i++){
			regionMeanCh1[i] = 0;
			regionMeanCh2[i] = 0;
			cellOutline[i] = null;
			
			//compute regionMean
			double[] watershedRegionMap = getSubImage(regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], labelMap.getWidth(), labelMapPixels);
			int size = watershedRegionMap.length;
			
			for(int j=0; j<size; j++){
				if((i+11)==(int)Math.round(watershedRegionMap[j])){
					regionMeanCh1[i] += rawDataCh1[i][j];
					regionMeanCh2[i] += rawDataCh2[i][j];
				}
			}
			regionMeanCh1[i] /= regionArea[i];
			regionMeanCh2[i] /= regionArea[i];
			
			FloatProcessor singleCellDataCh1 = new FloatProcessor(regionWidth[i], regionHeight[i], rawDataCh1[i]);
			FloatProcessor singleCellDataCh2 = new FloatProcessor(regionWidth[i], regionHeight[i], rawDataCh2[i]);

			if(!(regionXmin[i]<=1 || regionYmin[i]<=1 || regionXmax[i]>=(nx-2) || regionYmax[i]>=(ny-2))){
				//segment with snake-based approach
				Snake2DKeeper keeper = new Snake2DKeeper(); 
				cellOutline[i] = outlineCellWithSnake(keeper, i, regionXmin[i], regionYmin[i], singleCellDataCh1, singleCellDataCh2, segmentationType, segmentationQuality);
			}
			
			if(cellOutline[i]==null){
				cellOutline[i] = outlineCellWithThreshold(labelMapPixels, labelMap.getWidth(), i, regionXmin[i], regionWidth[i], regionYmin[i], regionHeight[i], rawDataCh1[i], rawDataCh2[i]);
				cellArea[i] = CTCellAnalyzer.getCellArea(cellOutline[i]);
				segmentationQuality[i] = CTCellAnalyzer.getCellAreaVSConvexHullArea(cellOutline[i], cellArea[i]);
				segmentationType[i] = CTExtractedData.REGION;
			}else{
				cellArea[i] = CTCellAnalyzer.getCellArea(cellOutline[i]);
			}

			cellMeanCh1[i] = CTCellAnalyzer.getCellMean(cellOutline[i], cellArea[i], singleCellDataCh1, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i]);
			cellMeanCh2[i] = CTCellAnalyzer.getCellMean(cellOutline[i], cellArea[i], singleCellDataCh2, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i]);
			cellTop3MeanCh1[i] = CTCellAnalyzer.getCellTopNMean(cellOutline[i], cellArea[i], singleCellDataCh1, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], 3);
			cellTop3MeanCh2[i] = CTCellAnalyzer.getCellTopNMean(cellOutline[i], cellArea[i], singleCellDataCh2, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], 3);
			cellTop5MeanCh1[i] = CTCellAnalyzer.getCellTopNMean(cellOutline[i], cellArea[i], singleCellDataCh1, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], 5);
			cellTop5MeanCh2[i] = CTCellAnalyzer.getCellTopNMean(cellOutline[i], cellArea[i], singleCellDataCh2, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], 5);
			cellMedianCh1[i] = CTCellAnalyzer.getCellMedian(cellOutline[i], cellArea[i], singleCellDataCh1, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i]);
			cellMedianCh2[i] = CTCellAnalyzer.getCellMedian(cellOutline[i], cellArea[i], singleCellDataCh2, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i]);
			cellMeanVScellStdCh1[i] = CTCellAnalyzer.getCellMeanVScellStd(cellOutline[i], cellArea[i], cellMeanCh1[i], singleCellDataCh1, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i]);
			cellMeanVScellStdCh2[i] = CTCellAnalyzer.getCellMeanVScellStd(cellOutline[i], cellArea[i], cellMeanCh1[i], singleCellDataCh2, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i]);
			cellCircularity[i] = CTCellAnalyzer.computeCircularity(cellOutline[i], cellArea[i]);
			minValueRegionCh1[i] = CTCellAnalyzer.getMinValueRegion(i, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], rawDataCh1[i], watershedRegionMap);
			minValueRegionCh2[i] = CTCellAnalyzer.getMinValueRegion(i, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], rawDataCh2[i], watershedRegionMap);
			minValueRegionNotCellCh1[i] = CTCellAnalyzer.getMinValueRegionNotCell(cellOutline[i], i, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], rawDataCh1[i], watershedRegionMap);
			minValueRegionNotCellCh2[i] = CTCellAnalyzer.getMinValueRegionNotCell(cellOutline[i], i, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], rawDataCh2[i], watershedRegionMap);
			regionBottomNPercentValueCh1[i] = CTCellAnalyzer.getRegionBottomNPercentValue(i, 5.0, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], rawDataCh1[i], watershedRegionMap);
			regionBottomNPercentValueCh2[i] = CTCellAnalyzer.getRegionBottomNPercentValue(i, 5.0, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], rawDataCh2[i], watershedRegionMap);
			regionNotCellBottomNPercentValueCh1[i] = CTCellAnalyzer.getRegionNotCellBottomNPercentValue(cellOutline[i], i, 5.0, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], rawDataCh1[i], watershedRegionMap);
			regionNotCellBottomNPercentValueCh2[i] = CTCellAnalyzer.getRegionNotCellBottomNPercentValue(cellOutline[i], i, 5.0, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], rawDataCh2[i], watershedRegionMap);
			regionNotCellMeanVSCellMeanCh1[i] = CTCellAnalyzer.getRegionNotCellMeanVSCellMean(cellOutline[i], cellMeanCh1[i], i, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], rawDataCh1[i], watershedRegionMap);
			regionNotCellMeanVSCellMeanCh2[i] = CTCellAnalyzer.getRegionNotCellMeanVSCellMean(cellOutline[i], cellMeanCh2[i], i, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], rawDataCh2[i], watershedRegionMap); 
			energyLoG10Ch1[i] = CTCellAnalyzer.getCellLoGEnergy(cellOutline[i], cellArea[i], singleCellDataCh1, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], 10.0, param[0]);
			energyLoG10Ch2[i] = CTCellAnalyzer.getCellLoGEnergy(cellOutline[i], cellArea[i], singleCellDataCh2, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], 10.0, param[0]);
			energyLoG3Ch1[i] = CTCellAnalyzer.getCellLoGEnergy(cellOutline[i], cellArea[i], singleCellDataCh1, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], 3.0, param[0]);
			energyLoG3Ch2[i] = CTCellAnalyzer.getCellLoGEnergy(cellOutline[i], cellArea[i], singleCellDataCh2, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], 3.0, param[0]);
			energySteerable4Ch1[i] = CTCellAnalyzer.getCellSteerable4Energy(cellOutline[i], cellArea[i], singleCellDataCh1, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], param[1]);
			energySteerable4Ch2[i] = CTCellAnalyzer.getCellSteerable4Energy(cellOutline[i], cellArea[i], singleCellDataCh2, regionXmin[i], regionYmin[i], regionWidth[i], regionHeight[i], param[1]);
		}
		cTExtractedData.setRegionNotCellMeanVSCellMeanCh1(regionNotCellMeanVSCellMeanCh1);
		cTExtractedData.setRegionNotCellMeanVSCellMeanCh2(regionNotCellMeanVSCellMeanCh2);
		cTExtractedData.setCellMeanVScellStdCh1(cellMeanVScellStdCh1);
		cTExtractedData.setCellMeanVScellStdCh2(cellMeanVScellStdCh2);
		cTExtractedData.setMinValueRegionCh1(minValueRegionCh1);
		cTExtractedData.setMinValueRegionCh2(minValueRegionCh2);
		cTExtractedData.setMinValueRegionNotCellCh1(minValueRegionNotCellCh1);
		cTExtractedData.setMinValueRegionNotCellCh2(minValueRegionNotCellCh2);
		cTExtractedData.setRegionBottomNPercentValueCh1(regionBottomNPercentValueCh1);
		cTExtractedData.setRegionBottomNPercentValueCh2(regionBottomNPercentValueCh2);
		cTExtractedData.setRegionNotCellBottomNPercentValueCh1(regionNotCellBottomNPercentValueCh1);
		cTExtractedData.setRegionNotCellBottomNPercentValueCh2(regionNotCellBottomNPercentValueCh2);
		cTExtractedData.setEnergyLoG10Ch1(energyLoG10Ch1);
		cTExtractedData.setEnergyLoG10Ch2(energyLoG10Ch2);
		cTExtractedData.setEnergyLoG3Ch1(energyLoG3Ch1);
		cTExtractedData.setEnergyLoG3Ch2(energyLoG3Ch2);
		cTExtractedData.setEnergySteerable4Ch1(energySteerable4Ch1);
		cTExtractedData.setEnergySteerable4Ch2(energySteerable4Ch2);
		cTExtractedData.setRegionMeanCh1(regionMeanCh1);
		cTExtractedData.setRegionMeanCh2(regionMeanCh2);
		cTExtractedData.setCellMeanCh1(cellMeanCh1);
		cTExtractedData.setCellMeanCh2(cellMeanCh2);
		cTExtractedData.setCellTop3MeanCh1(cellTop3MeanCh1);
		cTExtractedData.setCellTop3MeanCh2(cellTop3MeanCh2);
		cTExtractedData.setCellTop5MeanCh1(cellTop5MeanCh1);
		cTExtractedData.setCellTop5MeanCh2(cellTop5MeanCh2);
		cTExtractedData.setCellMedianCh1(cellMedianCh1);
		cTExtractedData.setCellMedianCh2(cellMedianCh2);
		cTExtractedData.setCellOutline(cellOutline);
		cTExtractedData.setCellArea(cellArea);
		cTExtractedData.setSegmentationType(segmentationType);
		cTExtractedData.setSegmentationQuality(segmentationQuality);
		cTExtractedData.setCellCircularity(cellCircularity);
	}
	
	private Polygon outlineCellWithSnake(Snake2DKeeper keeper, int i, int regionXmin, int regionYmin, FloatProcessor singleCellDataCh1, FloatProcessor singleCellDataCh2, int[] segmentationType, double[] segmentationQuality){
		
		int regionWidth =  singleCellDataCh1.getWidth();
		int regionHeight = singleCellDataCh1.getHeight();
		int maxLife = 200;
		Polygon cellOutline = null;
		
		double sigma = 2.5;
		int M = 5;
		double alpha = 1;
		int energyType = CTESnake.CONTOURENERGY;
		int detect = CTESnake.DETECTDARK;
		
		CTESnake eSnake = new CTESnake (singleCellDataCh1, sigma, maxLife, M, alpha, true, detect, energyType, null);
		keeper.optimize(eSnake, null);
		
		
		Snake2DScale[] skin = null;
		if(eSnake.getOrientation()==CTESnake.COUNTERCLOCKWISE && eSnake.getArea()<(regionWidth*regionHeight)){
			skin = eSnake.getScales();
			segmentationType[i] = CTExtractedData.ESNAKE;
			segmentationQuality[i] = eSnake.energy();
		}
		
		if(skin==null){
			return null;
		}
		
		int npoints   = skin[1].npoints;
		int[] xpoints = skin[1].xpoints;
		int[] ypoints = skin[1].ypoints;
		for(int point=0; point<npoints; point++){
			xpoints[point] += regionXmin;
			ypoints[point] += regionYmin;
		}
		cellOutline = new Polygon(xpoints, ypoints, npoints);
		
		return cellOutline;
	}
	
	
	private boolean[] labelBiggestComponent(boolean[] binaryMap, int regionWidth, int regionHeight){
		//labelize binary map
		int size = regionWidth*regionHeight;
		boolean[] T = new boolean[size];
		int regionWidth1 = regionWidth-1;
		CTUnsortedPairSortedList equivalenceClasses = new CTUnsortedPairSortedList();
		
		boolean LEFT, UP;
		int[] labelizedBinaryMap = new int [size];
		for(int i=0; i<size; i++){
			labelizedBinaryMap[i] = 0;
		}
		
		int nLabels = 0;
		
		//i=0 and j=0
		
		if(binaryMap[0]){
			nLabels++;
			labelizedBinaryMap[0] = nLabels;
		}
		
		//i=0 and j!=0
		for(int j=1; j<regionWidth; j++){
			if(binaryMap[j]){
				if (binaryMap[j-1]){
					labelizedBinaryMap[j] = labelizedBinaryMap[j-1]; 
				}else{
					nLabels++;
					labelizedBinaryMap[j] = nLabels;
				}
			}
		}		
		
		//i!=0 
		for(int i=1; i<regionHeight; i++){
			//j==0
			if(binaryMap[i*regionWidth]){
				if (binaryMap[(i-1)*regionWidth]){
					labelizedBinaryMap[i*regionWidth] = labelizedBinaryMap[(i-1)*regionWidth]; 
				}else{
					nLabels++;
					labelizedBinaryMap[i*regionWidth] = nLabels;
				}		
			}
			
			for(int j=0; j<regionWidth1; j++){
				if(binaryMap[j+i*regionWidth]){
					UP 	 = binaryMap[j+(i-1)*regionWidth];
					LEFT = binaryMap[(j-1)+i*regionWidth];
					if (UP && !LEFT){
						labelizedBinaryMap[j+i*regionWidth] = labelizedBinaryMap[j+(i-1)*regionWidth];
					}else if(!UP && LEFT){
						labelizedBinaryMap[j+i*regionWidth] = labelizedBinaryMap[j-1+i*regionWidth];
					}else if(!UP && !LEFT){
						nLabels++;
						labelizedBinaryMap[j+i*regionWidth] = nLabels;
					}else if(UP && LEFT){
						if(labelizedBinaryMap[j-1+i*regionWidth]==labelizedBinaryMap[j+(i-1)*regionWidth]){
							labelizedBinaryMap[j+i*regionWidth] = labelizedBinaryMap[j-1+i*regionWidth];
						}else{
							labelizedBinaryMap[j+i*regionWidth] = labelizedBinaryMap[j-1+i*regionWidth];
							equivalenceClasses.addElement(labelizedBinaryMap[j-1+i*regionWidth], labelizedBinaryMap[j+(i-1)*regionWidth]);
						}
					}
				}
			}
			
			//j==regionWidth-1
			if(binaryMap[regionWidth1+i*regionWidth]){
				if(binaryMap[regionWidth1-1+i*regionWidth]){
					labelizedBinaryMap[regionWidth1+i*regionWidth] = labelizedBinaryMap[regionWidth1-1+i*regionWidth];
				}else{
					nLabels++;
					labelizedBinaryMap[regionWidth1+i*regionWidth] = nLabels;
				}
			}
			
		}
		
		boolean[][] equivalenceClassesFlags = new boolean[nLabels+1][nLabels+1];
		for(int i=0; i<=nLabels; i++){
			for(int j=0; j<=nLabels; j++){
				equivalenceClassesFlags[i][j]=false;
			}
		}
		int nEquivalencePairs = equivalenceClasses.size();
		
		for(int i=0; i<=nLabels; i++){
			equivalenceClassesFlags[i][i]=true;
		}
		
		for(int i=0; i<nEquivalencePairs; i++){
			int min = equivalenceClasses.minElementAt(i);
			int max = equivalenceClasses.maxElementAt(i);
			equivalenceClassesFlags[min][max] = true;
			equivalenceClassesFlags[max][min] = true;
			
			for(int j=0; j<=nLabels; j++){
				if(equivalenceClassesFlags[min][j]){
					equivalenceClassesFlags[j][max] = true;
					equivalenceClassesFlags[max][j] = true;
				}
				
				if(equivalenceClassesFlags[max][j]){
					equivalenceClassesFlags[j][min] = true;
					equivalenceClassesFlags[min][j] = true;
				}
			}
		
		}

		int[] equivalenceClassesRepresentative = new int[nLabels+1];
		equivalenceClassesRepresentative[0] = 0;
		for(int i=1; i<=nLabels; i++){
			boolean found = false;
			for(int j=1; j<=nLabels && !found; j++){
				if(equivalenceClassesFlags[i][j]){
					found = true;
					equivalenceClassesRepresentative[i] = j;
				}
			}
		}
		
		for(int i=0; i<size; i++){
			labelizedBinaryMap[i]=equivalenceClassesRepresentative[labelizedBinaryMap[i]];
		}
		
		int[] nPixelsInClass = new int[nLabels+1];
		for(int i=0; i<=nLabels; i++){
			nPixelsInClass[i] = 0;
		}
		for(int i=0; i<size; i++){
			nPixelsInClass[labelizedBinaryMap[i]]++;
		}
		
		int biggerClass=1;
		for(int i=2; i<=nLabels; i++){
			if(nPixelsInClass[i]>nPixelsInClass[biggerClass]){
				biggerClass = i;
			}
		}
		
		for(int i=0; i<size; i++){
				T[i] = (labelizedBinaryMap[i]==biggerClass);
		}
	
		return T;
	}
	
	private Polygon outlineCellWithThreshold(float[] labelMapPixels, int labelMapWidth, int i, int regionXmin, int regionWidth, int regionYmin, int regionHeight, double[] rawDataCh1, double[] rawDataCh2){
		Polygon cellOutline = null;
		
		//segment with watershed-based approach
		double[] watershedRegionMap = getSubImage(regionXmin, regionYmin, regionWidth, regionHeight, labelMapWidth, labelMapPixels);
		boolean[] mask = generateMask(i, watershedRegionMap, Arrays.copyOf(rawDataCh1, regionWidth*regionHeight), regionWidth, regionHeight); 
		
		boolean[] T = labelBiggestComponent(mask, regionWidth, regionHeight);
		T = CTBinaryDilation.binaryDilation(T, regionWidth, regionHeight);
		CTContourTracing cTContourTracing = new CTContourTracing(T, regionWidth, regionHeight);
		cTContourTracing.trace();
		
		int[] xpoints = cTContourTracing.getXCoordinates();
		int[] ypoints = cTContourTracing.getYCoordinates();
		int npoints = cTContourTracing.getNPoints();
		
		//downsampling of the contour
		int reduction = 5;
		if(npoints>=4*reduction){
			int npoints2 = (int)Math.floor((double)npoints/(double)reduction);
			int[] xpoints2 = new int[npoints2];
			int[] ypoints2 = new int[npoints2];

			for(int cont=0; cont<npoints2; cont++ ){
				xpoints2[cont] = xpoints[(reduction*cont)%npoints];
				ypoints2[cont] = ypoints[(reduction*cont)%npoints];
			}

			for(int point=0; point<npoints2; point++){
				xpoints2[point] += regionXmin;
				ypoints2[point] += regionYmin;
			}
			cellOutline = new Polygon(xpoints2, ypoints2, npoints2);
		}else{
			for(int point=0; point<npoints; point++){
				xpoints[point] += regionXmin;
				ypoints[point] += regionYmin;
			}
			cellOutline = new Polygon(xpoints, ypoints, npoints);
		}
		return cellOutline;
	}

	private boolean[] generateMask(int nregion, double[] labelMapR, double[] imR, int wR, int hR){
		int size = wR*hR;
		boolean[] maskR = new boolean[size];
		double max = getMax(imR);
		
		for(int i=0; i<size; i++){
			if((int)labelMapR[i]!=(nregion+11)){
				imR[i] = max;
			}
		}
		
		double min = getMin(imR);
		int[] imRClasses = new int[size];
		
		for(int i=0; i<size; i++){
			imRClasses[i] = (int)Math.floor(255*(imR[i]-min)/(max-min));
		}
		
		double th = (max-min)*IJDefaultThreshold(getHistogram(imRClasses))/255.0 +min;
		
		for(int i=0; i<size; i++){
			if(imR[i]<th){
				maskR[i] = true;
			}else{
				maskR[i] = false;
			}
		}
		return(maskR);
	}
	
	private int[] getHistogram(int[] array){
		int[] histo = new int[256];
		int size = array.length;
		
		for(int i=0; i<size; i++){
			histo[array[i]]++;
		}
		return histo;
	}
	 
	private double[] getSubImage(int xmin, int ymin, int w, int h, int bigW, float[] im){
		double[] subim = new double[w*h];
		int xmax = xmin + w-1;
		int ymax = ymin + h-1;
		
		int ii = 0;
		for(int i=xmin; i<=xmax; i++){
			int jj=0;
			for(int j=ymin; j<=ymax; j++){
				double tmp = (double)im[i+j*bigW];
				subim[ii+jj*w] = tmp;
				jj++;
			}
			ii++;
		}
		return subim;
	}
	
	private double getMin(double[] array){
		double min = Double.POSITIVE_INFINITY;
		int size = array.length;
		for(int i=0; i<size; i++){
			if(array[i]<min) min = array[i]; 
		}
		return min;
	}
	
	private double getMax(double[] array){
		double max = Double.NEGATIVE_INFINITY;
		int size = array.length;
		for(int i=0; i<size; i++){
			if(array[i]>max) max = array[i]; 
		}
		return max;
	}
	
	private int IJDefaultThreshold(int [] data) {
		// Original IJ implementation for compatibility.
		int level;
		int maxValue = data.length - 1;
		double result, sum1, sum2, sum3, sum4;
		
		int min = 0;
		while ((data[min]==0) && (min<maxValue))
			min++;
		int max = maxValue;
		while ((data[max]==0) && (max>0))
			max--;
		if (min>=max) {
			level = data.length/2;
			return level;
		}
		
		int movingIndex = min;
		do {
			sum1=sum2=sum3=sum4=0.0;
			for (int i=min; i<=movingIndex; i++) {
				sum1 += i*data[i];
				sum2 += data[i];
			}
			for (int i=(movingIndex+1); i<=max; i++) {
				sum3 += i*data[i];
				sum4 += data[i];
			}			
			result = (sum1/sum2 + sum3/sum4)/2.0;
			movingIndex++;
		} while ((movingIndex+1)<=result && movingIndex<max-1);
		
		level = (int)Math.round(0.4*result);
		return level;
	}

	private void computeRegionMask(Roi roi){
		writeLog("Computing not moving region...");
		Rectangle roiBox = roi.getBounds();
		int width = roiBox.width;
		int height = roiBox.height;
		int totalSize = width*height;
		
		minVal = Integer.MAX_VALUE;
		maxVal = 0;
		double[] meanArray = new double[totalSize];
		double[] stdArray = new double[totalSize];
		regionMask = new boolean[totalSize];
		
		for(int i=0; i<totalSize; i++){
			meanArray[i] = 0;
			stdArray[i] = 0;
		}

		for(int frame=0; frame<Nstack; frame++){
			FloatProcessor ipr = CTInput.readProcessor(inputFolderCh1,fileListCh1[frame], roi);
			float[] currentFrame = (float[])ipr.getPixels();	
			
			for(int i=0; i<totalSize; i++){
				if(currentFrame[i]<minVal) minVal = (int)Math.round(currentFrame[i]);
				if(currentFrame[i]>maxVal) maxVal = (int)Math.round(currentFrame[i]);
				
				double oldMean = meanArray[i];
				meanArray[i] += (currentFrame[i]-meanArray[i])/(frame+1);
				stdArray[i] += (currentFrame[i]-meanArray[i])*(currentFrame[i]-oldMean);
			}
		}

		for(int i=0; i<totalSize; i++){
			stdArray[i] = Math.sqrt(stdArray[i]/(Nstack*(maxVal-minVal)));
			regionMask[i] = (stdArray[i]<runParameters.sigmaNotMovingRegion);
    	}
		
		if(showTempStd){
			writeLog("Showing temporal standard deviation");
			FloatProcessor tempStdProcessor = new FloatProcessor(width, height, stdArray);
			ImagePlus tempStdImagPlus = new ImagePlus("Normalized temporal standard deviation",tempStdProcessor);
			tempStdImagPlus.show();
		}
		
		if(showNMRMap){
			writeLog("Showing non-moving region map");
			float[] dataout = new float[totalSize];
			for(int i=0; i<totalSize; i++){
				if(regionMask[i]){
					dataout[i] = 255f;
				}else{
					dataout[i] = 0f;
				}
			}
			FloatProcessor NMRMapProcessor = new FloatProcessor(width, height, dataout,null);
			ImagePlus NMRMapImagPlus = new ImagePlus("Mask from temporal standard deviation",NMRMapProcessor);
			NMRMapImagPlus.show();
		}
	}

	public boolean isSuccessful() {
		return successful;
	}

	
}
